###################################################################################################
# zmssqldb.py
#
# $Id: zmssqldb.py,v 1.2 2003/07/10 09:55:07 nordmann Exp $
# $Name:  $
# $Author: nordmann $
# $Revision: 1.2 $
#
# Implementation of class ZMSSqlDb (see below).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###################################################################################################

# Imports.
from __future__ import nested_scopes
from Globals import HTML, HTMLFile
from Products.ZSQLMethods.SQL import SQLConnectionIDs
import copy
import sys
import urllib
# Product Imports.
from zmsobject import ZMSObject
import _globals
import _objattrs


SQLConnectionIDs = SQLConnectionIDs


###################################################################################################
###################################################################################################
###   
###   C o n s t r u c t o r ( s )
###   
###################################################################################################
###################################################################################################
def manage_addZMSSqlDb(self, lang, manage_lang, _sort_id, REQUEST):
  """ manage_addZMSSqlDb """
    
  ##### Create ####
  id_prefix = _globals.id_prefix(REQUEST.get('id','e'))
  obj = ZMSSqlDb(self.getNewId(id_prefix),_sort_id+1)
  self._setObject(obj.id, obj)
  
  obj = getattr(self,obj.id)
  ##### Object State ####
  obj.setObjStateNew(REQUEST)
  ##### VersionManager ####
  obj.onChangeObj(REQUEST)
            
  ##### Normalize Sort-IDs ####
  self.normalizeSortIds(id_prefix)
        
  # Return with message.        
  if REQUEST.RESPONSE:
    message = self.getLangStr('MSG_INSERTED',manage_lang)%obj.display_type(REQUEST)
    REQUEST.RESPONSE.redirect('%s/%s/manage_main?lang=%s&manage_lang=%s&manage_tabs_message=%s'%(self.absolute_url(),obj.id,lang,manage_lang,urllib.quote(message)))


###################################################################################################
###################################################################################################
###   
###   C l a s s
###   
###################################################################################################
###################################################################################################

class ZMSSqlDb(ZMSObject):

    # Properties.
    # -----------
    meta_type = "ZMSSqlDb"
    icon = "misc_/zms/zmssqldb.gif"
    icon_disabled = "misc_/zms/zmssqldb.gif"

    # Management Options.
    # -------------------
    manage_options = ( 
	{'label': 'TAB_EDIT',		'action': 'manage_main'},
	{'label': 'TAB_CONFIGURATION',	'action': 'manage_properties'},
	)

    # Management Permissions.
    # -----------------------
    __authorPermissions__ = (
		'manage','manage_main','manage_workspace',
		'manage_properties','manage_changeProperties',
		'manage_moveObjUp','manage_moveObjDown',
		'manage_cutObjects','manage_copyObjects','manage_pasteObjs',
		)
    __editorPermissions__ = (
		)
    __administratorPermissions__ = (
		)
    __ac_permissions__=(
		('ZMS Author', __authorPermissions__),
		('ZMS Editor', __editorPermissions__),
		('ZMS Administrator', __administratorPermissions__),
		)

    # Properties.
    # -----------
    __obj_attrs__ = {
        # Changed by
        'change_uid':{'datatype':'string','multilang':1,'xml':0,'lang_inherit':_globals.false},
        'change_dt':{'datatype':'datetime','multilang':1,'xml':0,'lang_inherit':_globals.false},
        # Properties
        'sort_id':{'datatype':'string','xml':0},
    }


    # Management Interface.
    # ---------------------
    manage_main = HTMLFile('dtml/zmssqldb/manage_main', globals())
    manage_properties = HTMLFile('dtml/zmssqldb/manage_properties', globals())


    """
    ###############################################################################################    
    #
    #   CONSTRUCTOR
    #
    ###############################################################################################    
    """

    # -------------------------------------------------------------------------------------------------
    # ZMSSqlDb.recurse_updateVersion:
    #
    # Update version.
    # -------------------------------------------------------------------------------------------------
    def recurse_updateVersion(self, REQUEST):
      message = ''

      # Process super-objects.
      message = message + ZMSObject.recurse_updateVersion(self,REQUEST)
        
      # Process object.
      # [...]

      # Return with message.
      return message


    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.sql_quote__:
    # ------------------------------------------------------------------------------------------
    def sql_quote__(self, tablename, key, v):
      attr = self.getEntityAttr(tablename,key)
      if attr['type'] in ['float','int']:
        return "%s" % str(v)
      else:
        v = str(v)
        if v.find("\'") >= 0: v=''.join(v.split("\'"),"''")
        return "'%s'"%v

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.query:
    # ------------------------------------------------------------------------------------------
    def query(self, qs):
      from cStringIO import StringIO
      from Shared.DC.ZRDB.Results import Results
      from Shared.DC.ZRDB import RDB
      databaseAdptr = getattr(self,self.connection_id)
      dbc = databaseAdptr._v_database_connection
      try:
        res = dbc.query(qs)
        if type(res) is type(''):
          f=StringIO()
          f.write(res)
          f.seek(0)
          result=RDB.File(f)
        else:
          result=Results(res)
      except:
        result = '[Exception in %s.query]:<br>'%(self.meta_type)
        result = result + 'Type: <b>%s</b><br>'%str(sys.exc_type)
        result = result + 'Value: <b>%s</b><br>'%str(sys.exc_value)
        result = result + 'Traceback: <b>%s</b><br>'%str(sys.exc_traceback)
        result = result + '<code>%s</code><br>'%qs
      return result

    
    """
    ###############################################################################################
    #
    #   E N T I T I E S
    #
    ###############################################################################################
    """

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.getEntities:
    # 
    # Returns dictionary of entities.
    # ------------------------------------------------------------------------------------------
    def getEntities(self):
      return getattr(self,'entities',{})

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.getEntityIds:
    # 
    # Returns list of Ids of entities.
    # ------------------------------------------------------------------------------------------
    def getEntityIds(self, sort=1):
      entities = self.getEntities()
      ids = entities.keys()
      if sort:
        mapping = map(lambda id: (entities[id]['name'],id),ids)
        mapping.sort()
        ids = map(lambda x: x[1],mapping)
      return ids

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.getEntity:
    # 
    # Returns Meta-Object identified by ID.
    # ------------------------------------------------------------------------------------------
    def getEntity(self, id):
      entity = self.getEntities()[id]
      name = entity['name']
      attrs = entity['attrs']
      return {'id':id,'name':name,'attrs':attrs}

    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.initEntity:
    # 
    #  Init entity with specified values.
    # ------------------------------------------------------------------------------------------
    def initEntity(self, id, entity):
      entities = self.getEntities()
      entities[id] = entity
      self.entities = entities.copy()

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.setEntity:
    # 
    # Set/add entity with specified values.
    # ------------------------------------------------------------------------------------------
    def setEntity(self, newId, newName, attrs=[]):
      entities = self.getEntities()
      entity = {}
      entity['name'] = newName
      entity['attrs'] = copy.copy(attrs)
      self.initEntity(newId,entity)

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.delEntity:
    # 
    # Delete entity specified by Id.
    # ------------------------------------------------------------------------------------------
    def delEntity(self, id):
      cp = self.getEntities()
      entities = {}
      for key in cp.keys():
        if key != id:
          entities[key] = cp[key]
      self.entities = entities.copy()



    """
    ###############################################################################################
    #
    #   A T T R I B U T E S
    #
    ###############################################################################################
    """

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.initEntityAttrs:
    # 
    # Init list of attributes for entity specified by id.
    # ------------------------------------------------------------------------------------------
    def initEntityAttrs(self, id):
      entities = self.getEntities()
      entity = entities[id]
      entity['attrs'] = []
      entities[id] = entity
      self.entities = entities.copy()

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.getEntityAttrIds:
    # 
    # Returns list of attribute-Ids for entity specified by id.
    # ------------------------------------------------------------------------------------------
    def getEntityAttrIds(self, id):
      ids = []
      entity = self.getEntities()[id]
      attrs = entity['attrs']
      for attr in attrs:
        ids.append(attr['id'])
      return ids

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.getEntityAttr:
    # 
    # Get attribute for entity specified by id.
    # ------------------------------------------------------------------------------------------
    def getEntityAttr(self, id, key):
      entity = self.getEntities()[id]
      attrs = entity['attrs']
      for attr in attrs:
        if attr['id'] == key:
          attr = attr.copy()
          attr['key'] = key
          attr['mandatory'] = attr.get('mandatory',0)
          attr['size'] = attr.get('size',35)
          attr = _objattrs.syncobjattr(self,attr)
          return attr
      return None

    # ------------------------------------------------------------------------------------------
    # ZMSSqlDb.setEntityAttr:
    # 
    # Set/add attribute for entity with specified values.
    # ------------------------------------------------------------------------------------------
    def setEntityAttr(self, id, newId, newName, newPk=0, newMandatory=0, newChangeOnInsert=1, newChangeOnUpdate=1, newType='string', newKeys=[]):
      entities = self.getEntities()
      entity = entities[id]
      attrs = copy.copy(entity['attrs'])
      attr = {}
      attr['id'] = newId
      attr['name'] = newName
      attr['pk'] = newPk
      attr['mandatory'] = newMandatory or newPk
      attr['change_on_insert'] = newChangeOnInsert
      attr['change_on_update'] = newChangeOnUpdate
      attr['type'] = newType
      attr['keys'] = newKeys
      ids = self.getEntityAttrIds(id)
      if newId in ids:
        attrs[ids.index(newId)] = attr
      else:
        attrs.append(attr)
      entity['attrs'] = attrs
      entities[id] = entity
      self.entities = entities.copy()


    """
    ###############################################################################################    
    ###
    ###   P r o p e r t i e s
    ###
    ###############################################################################################    
    """

    ###############################################################################################
    #  ZMSSqlDb.manage_changeProperties: 
    #
    #  Change Sql-Database properties.
    ###############################################################################################
    def manage_changeProperties(self, lang, manage_lang, btn='', key='all', REQUEST=None, RESPONSE=None): 
      """ ZMSSqlDb.manage_changeProperties """
      
      message = ''
      id = REQUEST.get('id','')
      
      # Synchronize.
      # ------------
        
      if btn == self.getLangStr('BTN_SYNCHRONIZE',manage_lang):
        entities = {}
        databaseAdptr = getattr(self,self.connection_id)
        for tableBrwsr in databaseAdptr.tpValues():
          tableName = tableBrwsr.Name().lower()
          tableType = tableBrwsr.Type().upper()
          if tableType == 'TABLE':
            entity = {}
            entity['name'] = tableName.capitalize().replace('_',' ')
            entity['attrs'] = []
            for columnBrwsr in tableBrwsr.tpValues():
              colId = columnBrwsr.tpId().lower()
              colDescr = columnBrwsr.Description().upper()
              if colDescr.find('INTEGER')>=0:
                colType = 'int'
              elif colDescr.find('CHAR')>=0:
                colSize = 50
                i = colDescr.find('(')
                if i >= 0:
                  j = colDescr.find(')')
                  if j >= 0:
                    colSize = int(colDescr[i+1:j])
                if colSize > 50:
                  colType = 'text'
                else:
                  colType = 'string'
              else:
                colType = 'string'
              #++ print tableName, colId, colDescr
              attrs = {}
              attrs['id'] = colId
              attrs['name'] = colId.capitalize().replace('_',' ')
              attrs['mandatory'] = colDescr.find('NOT NULL')>0
              attrs['type'] = colType
              attrs['keys'] = []
              entity['attrs'].append(attrs)
            entities[tableName] = entity
        self.entities = entities.copy()

      # Delete.
      # -------
        
      # Obj.
      elif btn == self.getLangStr('BTN_DELETE',manage_lang) and key == 'obj':
        self.delEntity(id)
        id = ''
        message = self.getLangStr('MSG_CHANGED',manage_lang)
        
      # Attr.
      elif btn == 'delete' and key == 'attr':
        pass
          
      # Change.
      # -------
      elif btn == self.getLangStr('BTN_CHANGE',manage_lang):
      
        # Connection Id.
        if REQUEST.has_key('connection_id'):
          self.connection_id = REQUEST['connection_id']
      
        # Obj.
        if key == 'all':
          id = REQUEST['id'].strip()
          name = REQUEST['obj_name'].strip()
          entity = self.getEntities()[id]
          attrs = entity['attrs']
          self.setEntity(id,name,attrs)

        # Attr.
        if key == 'all':
          saved_ids = self.getEntityAttrIds(id)
          self.initEntityAttrs(id)
          for saved_id in saved_ids:
            attr_id = REQUEST['attr_id_%s'%saved_id].strip()
            newName = REQUEST['attr_name_%s'%saved_id].strip()
            newPk = REQUEST.get('attr_pk_%s'%saved_id,0)
            newMandatory = REQUEST.get('attr_mandatory_%s'%saved_id,0)
            newChangeOnInsert = REQUEST.get('attr_change_on_insert_%s'%saved_id,0)
            newChangeOnUpdate = REQUEST.get('attr_change_on_update_%s'%saved_id,0)
            newType = REQUEST['attr_type_%s'%saved_id]
            newKeys = self.string_list(REQUEST.get('attr_keys_%s'%saved_id,''),'\n')
            self.setEntityAttr(id,attr_id,newName,newPk,newMandatory,newChangeOnInsert,newChangeOnUpdate,newType,newKeys)
          
        message = self.getLangStr('MSG_CHANGED',manage_lang)
        
        
      # Insert.
      # -------
      elif btn == self.getLangStr('BTN_INSERT',manage_lang):
      
        # Obj.
        if key == 'obj':
          newId = REQUEST['newId'].strip()
          newName = REQUEST['newName'].strip()
          self.setEntity(newId,newName)
          id = newId
          
        # Attr.
        elif key == 'attr':
          newId = REQUEST['attr_id'].strip()
          newName = REQUEST['attr_name'].strip()
          newPk = REQUEST.get('attr_pk',0)
          newMandatory = REQUEST.get('attr_mandatory',0)
          newChangeOnInsert = REQUEST.get('attr_change_on_insert',0)
          newChangeOnUpdate = REQUEST.get('attr_change_on_update',0)
          newType = REQUEST['attr_type']
          newKeys = self.string_list(REQUEST.get('attr_keys',''),'\n')
          self.setEntityAttr(id,newId,newName,newPk,newMandatory,newChangeOnInsert,newChangeOnUpdate,newType,newKeys)
      
      # Return with message.
      message = urllib.quote(message)
      return RESPONSE.redirect('manage_properties?lang=%s&manage_lang=%s&manage_tabs_message=%s&id=%s&attr_id=%s'%(lang,manage_lang,message,id,REQUEST.get('attr_id','')))


    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.isPage
    # ---------------------------------------------------------------------------------------------
    def isPage(self): return 0

    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.isPageElement
    # ---------------------------------------------------------------------------------------------
    def isPageElement(self): return 0

    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.getTitlealt
    # ---------------------------------------------------------------------------------------------
    def getTitlealt(self,REQUEST): return self.display_type(REQUEST)


    """
    ###############################################################################################
    ###
    ###  H T M L - P r e s e n t a t i o n 
    ###
    ###############################################################################################
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSSqlDb._getBodyContent:
    #
    #	HTML presentation of Textarea. 
    # ---------------------------------------------------------------------------------------------
    def _getBodyContent(self, REQUEST, thumbnail=0):
      # Render HTML presentation.
      text = '<h2>%s</h2>'%self.display_type(REQUEST)
      # Return <html>-presentation.
      return text

    # ---------------------------------------------------------------------------------------------
    #	ZMSSqlDb.renderShort:
    #
    #	Renders short presentation of Textarea.
    # ---------------------------------------------------------------------------------------------
    def renderShort(self, REQUEST):
      return self._getBodyContent(REQUEST,thumbnail=1)

###################################################################################################
